"""windowing functions for shape manipulations of images and frames."""

from typing import Literal, Optional, Tuple

import numpy as np
import psutil


def round_to_even(input_tuple: Tuple[int]):
    """Round tuple int values to ceil of nearest even number."""
    return tuple((x + 1) if x % 2 != 0 else x for x in input_tuple)


def available_memory(safety=0.0):
    """Get available memory in bytes."""
    return psutil.virtual_memory().free - safety  #  + psutil.swap_memory().free


def required_memory(n_frames, dim_size, window_size, overlap, search_area_size, safety=0.0):
    """Estimate the required amount of memory to hold subwindows and correlation results in memory."""
    win_shape = get_array_shape(
        dim_size=dim_size, window_size=window_size, overlap=overlap, search_area_size=search_area_size
    )
    # we need to store the sliced images, and sliced images correlations (which use one frame less)
    return ((2 * n_frames - 1) * win_shape[0] * win_shape[1] * search_area_size[0] * search_area_size[1] * 4) * (
        1 + safety
    )


def sliding_window_idx(
    image: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (32, 32),
    search_area_size: Optional[Tuple[int, int]] = None,
) -> np.ndarray:
    """Create y and x indices per interrogation window.

    Parameters
    ----------
    image : np.ndarray
        black-white image or template
    window_size : tuple
        size of interrogation window (y, x)
    overlap : tuple
        overlap of pixels of interrogation windows (y, x)
    search_area_size : Tuple[int, int], optional
        size of search area window (y, x), if not set, made equal to window_size

    Returns
    -------
    win_x : np.ndarray (w * y * x)
        x-indices of interrogation windows (w)
    win_y : np.ndarray (w * y * x)
        y-indices of interrogation windows (w)


    """
    search_area_size = window_size if search_area_size is None else search_area_size
    x, y = get_rect_coordinates(
        image.shape, window_size, overlap, search_area_size=search_area_size, center_on_field=False
    )
    xi, yi = np.meshgrid(x, y)
    xi = (xi - search_area_size[1] // 2).astype(int)
    yi = (yi - search_area_size[0] // 2).astype(int)
    xi, yi = np.reshape(xi, (-1, 1, 1)), np.reshape(yi, (-1, 1, 1))

    # make ranges of search area size
    win_x, win_y = np.meshgrid(np.arange(0, search_area_size[1]), np.arange(0, search_area_size[0]))
    # add center coordinates to search areas
    win_x = win_x[np.newaxis, :, :] + xi
    win_y = win_y[np.newaxis, :, :] + yi
    return win_x, win_y


def sliding_window_array(
    image: np.ndarray,
    win_x: np.ndarray,
    win_y: np.ndarray,
) -> np.ndarray:
    """Select a interrogation window from an image."""
    return image[win_y, win_x]


def multi_sliding_window_array(
    imgs: np.ndarray, win_x: np.ndarray, win_y: np.ndarray, swap_time_dim=False
) -> np.ndarray:
    """Select interrogation windows from a set of images."""
    windows = np.stack([sliding_window_array(img, win_x, win_y) for img in imgs])
    if swap_time_dim:
        return np.swapaxes(windows, 0, 1)
    return windows


def get_axis_shape(
    dim_size: int,
    window_size: int,
    overlap: int,
    search_area_size: Optional[int] = None,
) -> int:
    """Get shape of image axis given its dimension size.

    Parameters
    ----------
    dim_size : int
        size of axis [pix]
    window_size : int
        size of interrogation window over axis dimension [pix]
    overlap : int
        size of overlap [pix]
    search_area_size : int, optional
        size of search area window over axis dimension [pix]

    Returns
    -------
    int
        amount of interrogation windows over provided axis

    """
    # set the search area size to window size if not provided
    search_area_size = window_size if search_area_size is None else search_area_size
    axis_shape = (dim_size - search_area_size) // (window_size - overlap) + 1
    return axis_shape


def get_array_shape(
    dim_size: Tuple[int, int],
    window_size: Tuple[int, int],
    overlap: Tuple[int, int],
    search_area_size: Optional[Tuple[int, int]] = None,
):
    """Get the resulting shape of velocimetry results as a tuple of dimension sizes.

    Parameters
    ----------
    dim_size : [int, int]
        sizes of axes [pix]
    window_size : [int, int]
        sizes of interrogation windows [pix]
    overlap : [int, int]
        sizes of overlaps [pix]
    search_area_size : [int, int]
        size of search area windows [pix]

    Returns
    -------
    shape of array returned from velocimetry analysis

    """
    search_area_size = window_size if search_area_size is None else search_area_size
    array_shape = tuple(
        get_axis_shape(dim_size_, window_size_, overlap_)
        for (dim_size_, window_size_, overlap_, search_area_size_) in zip(
            dim_size, window_size, overlap, search_area_size
        )
    )
    return array_shape


def get_axis_coords(
    dim_size: int,
    window_size: int,
    overlap: int,
    search_area_size: Optional[int] = None,
    center_on_field: bool = False,
):
    """Get axis coordinates for one axis with provided dimensions and window size parameters.

    Overlap for windows can be provided.

    Parameters
    ----------
    dim_size : int
        size of axis [pix]
    window_size : int
        size of interrogation window over axis dimension [pix]
    overlap : int
        size of overlap [pix]
    search_area_size : int, optional
        size of search area window over axis dimension [pix], if not set, made equal to window_size
    center_on_field : bool, optional
        take the center of the window as coordinate (default, False)

    Returns
    -------
    x- or y-coordinates of resulting velocimetry grid

    """
    search_area_size = window_size if search_area_size is None else search_area_size
    # get the amount of expected coordinates
    ax_shape = get_axis_shape(dim_size, window_size, overlap, search_area_size)
    coords = np.arange(ax_shape) * (window_size - overlap) + (search_area_size) / 2.0
    if center_on_field is True:
        coords_shape = get_axis_shape(dim_size=dim_size, window_size=window_size, overlap=overlap)
        coords += (dim_size - 1 - ((coords_shape - 1) * (window_size - overlap) + (window_size - 1))) // 2
    return np.int64(coords)


def get_rect_coordinates(
    dim_size: Tuple[int, int],
    window_size: Tuple[int, int],
    overlap: Tuple[int, int],
    search_area_size: Optional[Tuple[int, int]] = None,
    center_on_field: bool = False,
):
    """Create coordinates (x, y) of velocimetry results.

    Overlap can be provided in case each interrogation window is to overlap with the neighbouring interrogation window.

    Parameters
    ----------
    dim_size : [int, int]
        sizes of axes [pix]
    window_size : [int, int]
        sizes of interrogation windows [pix]
    overlap : [int, int]
        sizes of overlaps [pix]
    search_area_size : [int, int], optional
        sizes of search area windows [pix], if not set, made equal to window_size
    center_on_field : bool, optional
        take the center of the window as coordinate (default, False)

    Returns
    -------
    x, y: np.ndarray (1D), np.ndarray (1D)
        x- and y-coordinates in axis form

    """
    search_area_size = window_size if search_area_size is None else search_area_size
    y = get_axis_coords(
        dim_size[0], window_size[0], overlap[0], search_area_size=search_area_size[0], center_on_field=center_on_field
    )
    x = get_axis_coords(
        dim_size[1], window_size[1], overlap[1], search_area_size=search_area_size[1], center_on_field=center_on_field
    )

    return x, y


def mask_search_area(window_size: Tuple[int, int] = (64, 64), search_area_size: Tuple[int, int] = (128, 128)):
    """Create mask to be used on first frame to only compare values inside the window size of the first frame.

    If either one of the dimensions in search_area_size is smaller than the same dimension in window size, a mask
    equal to window_size will be returned with only ones

    Parameters
    ----------
    window_size : [int, int]
        sizes of interrogation windows [pix]
    search_area_size : [int, int]
        sizes of the search area (includes interrogation window and area outside of that

    Returns
    -------
    np.ndarray
        mask, zeros in edge of search_area_size, ones in part of window_size

    """
    assert np.all(
        np.array(search_area_size) >= np.array(window_size)
    ), "At least one dimension of search_area_size is smaller than the same dimension of window_size"
    mask = np.zeros(search_area_size)
    pady = int((search_area_size[0] - window_size[0]) / 2)
    padx = int((search_area_size[1] - window_size[1]) / 2)
    mask[slice(pady, search_area_size[0] - pady), slice(padx, search_area_size[1] - padx)] = 1
    return mask


def normalize(imgs: np.ndarray, mode: Literal["xy", "time"] = "time"):
    """Normalize images assuming the last two dimensions contain the x/y image intensities.

    Parameters
    ----------
    imgs : np.ndarray (n x Y x X) or (n x m x Y x X)
        input images, organized in at least one stack
    mode : str, optional
        can be "xy" or "time" (default). manner over which normalization should be done, using time or space as
        dimension to normalize over.

    Returns
    -------
    imgs_norm : np.ndarray (n x Y x X) or (n x m x Y x X)
        output normalized images, organized in at least one stack, similar to imgs.

    """
    # compute means and stds
    if mode == "xy":
        imgs_std = np.expand_dims(imgs.reshape(imgs.shape[0], imgs.shape[1], -1).std(axis=-1), axis=(-1, -2))
        imgs_mean = np.expand_dims(imgs.reshape(imgs.shape[0], imgs.shape[1], -1).mean(axis=-1), axis=(-1, -2))
    elif mode == "time":
        imgs_std = np.expand_dims(imgs.std(axis=-3), axis=-3)
        imgs_mean = np.expand_dims(imgs.mean(axis=-3), axis=-3)
    else:
        raise ValueError(f'mode must be "xy" or "time", but is "{mode}"')
    img_norm = np.divide(imgs - imgs_mean, imgs_std, out=np.zeros_like(imgs, dtype=imgs.dtype), where=(imgs_std != 0))
    # img_norm = (imgs - imgs_mean) # / imgs_std
    # img_norm[imgs_std == 0] = 0
    return np.clip(img_norm, 0, img_norm.max())



"""pyFFTW cross-correlation related functions."""

import atexit
import multiprocessing
import os
import pickle

import numba as nb
import numpy as np
import pyfftw

from ffpiv import _WISDOM_FILE

# Configure pyFFTW for optimal performance
pyfftw.interfaces.cache.enable()
pyfftw.config.NUM_THREADS = multiprocessing.cpu_count()
pyfftw.config.PLANNER_EFFORT = "FFTW_MEASURE"


def _load_wisdom():
    """Load FFTW wisdom from file if available."""
    if os.path.exists(_WISDOM_FILE):
        try:
            with open(_WISDOM_FILE, "rb") as f:
                wisdom = pickle.load(f)
            pyfftw.import_wisdom(wisdom)
        except Exception:
            pass  # Silently ignore if we can't load wisdom


def _save_wisdom():
    """Save FFTW wisdom to file for future sessions."""
    try:
        wisdom = pyfftw.export_wisdom()
        with open(_WISDOM_FILE, "wb") as f:
            pickle.dump(wisdom, f)
    except Exception:
        pass  # Silently ignore if we can't save wisdom


_load_wisdom()

# Register save_wisdom to run at interpreter exit
atexit.register(_save_wisdom)


@nb.jit(nb.float32[:, :, :](nb.float32[:, :, :]), cache=True, nopython=True, parallel=True, nogil=True)
def normalize_intensity(img: np.ndarray) -> np.ndarray:
    """Normalize intensity of an image splitted in interrogation windows using numba back-end.

    Parameters
    ----------
    img : np.ndarray (y, x)
        Image window

    Returns
    -------
    np.ndarray
        (y, x) array with normalized intensities of window

    """
    for n in nb.prange(img.shape[0]):
        img_ = img[n]
        img_mean = np.float32(np.mean(img_))
        img_ = img_ - img_mean
        img_std = np.float32(np.std(img_))
        if img_std != 0:
            img_ = img_ / img_std
        else:
            img_ = np.zeros_like(img_, dtype=np.float32)
        img[n] = img_
    return img


@nb.jit(nb.float32[:, :, :](nb.float32[:, :, :]), cache=True, nopython=True, parallel=True, nogil=True)
def normalize_intensity_clip(img: np.ndarray) -> np.ndarray:
    """Normalize intensity and clip values to [0, max] image interrogation windows using numba back-end.

    Parameters
    ----------
    img : np.ndarray (w, y, x)
        Image splitted in interrogation windows (w)

    Returns
    -------
    np.ndarray
        (w, y, x) array with normalized intensities of all windows

    """
    for n in nb.prange(img.shape[0]):
        img_ = img[n]
        img_mean = np.float32(np.mean(img_))
        img_ = img_ - img_mean
        img_std = np.float32(np.std(img_))
        if img_std != 0:
            img_ = img_ / img_std
        else:
            img_ = np.zeros_like(img_, dtype=np.float32)
        img[n] = np.clip(img_, 0, img_.max())
    return img


nb.jit(nb.float32[:, :, :, :](nb.float32[:, :, :, :]), cache=True, nopython=True, parallel=True, nogil=True)


def multi_normalize_intensity(imgs: np.ndarray) -> np.ndarray:
    """Normalize intensities of several images in one go using numba back-end."""
    for m in nb.prange(imgs.shape[0]):
        imgs[m] = normalize_intensity(imgs[m])
    return imgs


nb.jit(nb.float32[:, :, :, :](nb.float32[:, :, :, :]), cache=True, nopython=True, parallel=True, nogil=True)


def multi_normalize_intensity_clip(imgs: np.ndarray) -> np.ndarray:
    """Normalize intensities of several images in one go and clip values to [0, max] using numba back-end."""
    for m in nb.prange(imgs.shape[0]):
        imgs[m] = normalize_intensity_clip(imgs[m])
    return imgs


def rfft2(x):
    """Compute 2D real FFT using pyFFTW.

    Parameters
    ----------
    x : np.ndarray
        Input array.

    Returns
    -------
    np.ndarray
        Complex array containing the FFT result.

    """
    x_align = pyfftw.empty_aligned(x.shape, dtype=np.float32)
    rfft = pyfftw.builders.rfft2(
        x_align, axes=(-2, -1), threads=multiprocessing.cpu_count(), planner_effort="FFTW_MEASURE"
    )
    rfft.input_array[:] = x
    return rfft()


def irfft2(x):
    """Compute 2D inverse real FFT using pyFFTW.

    Parameters
    ----------
    x : np.ndarray
        Input complex array.

    Returns
    -------
    np.ndarray
        Real array containing the inverse FFT result.

    """
    x_align = pyfftw.empty_aligned(x.shape, dtype=np.complex64)
    irfft = pyfftw.builders.irfft2(
        x_align, axes=(-2, -1), threads=multiprocessing.cpu_count(), planner_effort="FFTW_MEASURE"
    )  # , s=s
    irfft.input_array[:] = x
    return irfft()


def fftshift(x, axes=None):
    """Shift the zero-frequency component to the center using pyfftw's numpy interfacing.

    Parameters
    ----------
    x : np.ndarray
        Input array.
    axes : int or tuple of ints, optional
        Axes over which to shift.

    Returns
    -------
    np.ndarray
        Shifted array.

    """
    return pyfftw.interfaces.numpy_fft.fftshift(x, axes=axes)


def _ncc(image_a, image_b, norm=True, clip_norm=False, rfft2_builder=None, irfft2_builder=None):
    """Perform normalized cross-correlation on a set of interrogation window pairs with pyFFTW back-end.

    Parameters
    ----------
    image_a : np.ndarray
        uint8 type array [w, y, x] containing a single image, sliced into interrogation windows (w)
    image_b : np.ndarray
        uint8 type array [w, y, x] containing the next image, sliced into interrogation windows (w)
    norm : bool, optional
        If set, images will be normalized to have zero mean and 1 variance, if not, normalization should usually already
        be performed outside of this function.
    clip_norm : bool, optional
        If set to True, the normalized intensities are clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.
    rfft2_builder : pyfftw builder object, optional
        Pre-configured rfft2 builder for reuse. If None, creates a new one.
    irfft2_builder : pyfftw builder object, optional
        Pre-configured irfft2 builder for reuse. If rfft2_builder is provided, this builder must also be provided.

    Returns
    -------
    np.ndarray
        float64 [w * y * x] correlations of interrogation window pixels

    """
    const = np.multiply(*image_a.shape[-2:])
    if norm:
        if clip_norm:
            image_a = normalize_intensity_clip(image_a)
            image_b = normalize_intensity_clip(image_b)
        else:
            image_a = normalize_intensity(image_a)
            image_b = normalize_intensity(image_b)

    if rfft2_builder is not None and irfft2_builder is not None:
        # Use pre-configured builders
        rfft2_builder.input_array[:] = image_a
        f2a = np.conj(rfft2_builder())

        rfft2_builder.input_array[:] = image_b
        f2b = rfft2_builder()

        irfft2_builder.input_array[:] = f2a * f2b
        corr = irfft2_builder()

    else:
        f2a = np.conj(rfft2(image_a))
        f2b = rfft2(image_b)
        corr = irfft2(f2a * f2b)
    return np.clip(fftshift(corr.real, axes=(-2, -1)) / const, 0, 1)


def multi_img_ncc(imgs, mask=None, idx=None, clip_norm=False):
    """Compute correlation over all image pairs in `imgs` using pyFFTW back-end.

    Correlations are computed for each interrogation window (dim1) and each image pair (dim0)
    Because pair-wise correlation is performed the resulting dim0 size one stride smaller than the input imgs array.

    Parameters
    ----------
    imgs : np.ndarray
        [i, w, y, x] set of images (i), subdivided into windows (w) for cross-correlation computation.
    mask : np.ndarray
        [y, x] array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.
    clip_norm : bool, optional
        If set to True, the normalized intensities are clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.

    Returns
    -------
    np.ndarray
        float32 [(i - 1), w, y, x] correlations of interrogation window pixels for each image pair spanning i.

    """
    corr = np.empty((len(imgs) - 1, imgs.shape[-3], imgs.shape[-2], imgs.shape[-1]), dtype=np.float32)
    corr.fill(np.nan)
    if idx is None:
        idx = np.repeat(True, imgs.shape[-3])
    # imgs = imgs[:, idx, :, :]
    if clip_norm:
        imgs = multi_normalize_intensity_clip(imgs[:, idx, :, :])
    else:
        imgs = multi_normalize_intensity(imgs[:, idx, :, :])

    # Create aligned arrays and builders once, based on the indexed shape
    n_windows = idx.sum()
    window_shape = (n_windows, imgs.shape[-2], imgs.shape[-1])
    irfft2_input_shape = (n_windows, imgs.shape[-2], imgs.shape[-1] // 2 + 1)

    # rfft2 input: real array, output: complex array with shape[-1] // 2 + 1
    rfft2_input = pyfftw.empty_aligned(window_shape, dtype="float32", n=pyfftw.simd_alignment)
    rfft2_builder = pyfftw.builders.rfft2(
        rfft2_input,
        axes=(-2, -1),
        threads=multiprocessing.cpu_count(),
        planner_effort="FFTW_MEASURE",
        auto_align_input=False,
    )

    # irfft2 input: complex array (output shape of rfft2), output: real array
    irfft2_input = pyfftw.empty_aligned(irfft2_input_shape, dtype="complex64", n=pyfftw.simd_alignment)
    irfft2_builder = pyfftw.builders.irfft2(
        irfft2_input,
        axes=(-2, -1),
        threads=multiprocessing.cpu_count(),
        planner_effort="FFTW_MEASURE",
        auto_align_input=False,
    )
    # Pre-allocate temporary arrays for intermediate results (also aligned)
    for n in range(len(imgs) - 1):
        img_a = imgs[n] * mask[idx]
        img_b = imgs[n + 1]
        res = _ncc(
            image_a=img_a,
            image_b=img_b,
            norm=False,
            rfft2_builder=rfft2_builder,
            irfft2_builder=irfft2_builder,
        )
        corr[n, idx] = res
    return corr



"""FF-PIV: Fast and Flexible Particle Image Velocimetry analysis powered by numba."""

__version__ = "0.2.0"

from pathlib import Path

import platformdirs

from . import pnp, sample_data, window

try:
    import rocket_fft  # noqa: F401

    from . import pnb

    HAS_ROCKET_FFT = True
except ImportError:
    HAS_ROCKET_FFT = False

cache_dir = Path(platformdirs.user_cache_dir("ffpiv"))
if not cache_dir.exists():
    cache_dir.mkdir(parents=True)

_WISDOM_FILE = cache_dir / ".ffpiv_wisdom.pkl"

# import user top functionality
from .api import *



import numba as nb
import numpy as np


@nb.jit(parallel=True, cache=True, nopython=True)
def u_v_displacement(
    corr,
    n_rows,
    n_cols,
):
    """Compute u (x-direction) and v (y-direction) displacements.

    u and v displacements are computed from correlations in windows and number and rows / columns using numba
    back-end.

    Parameters
    ----------
    corr : np.ndarray
        (w, y, x) correlation planes for each interrogation window (w).
    n_rows : int
        number of rows in the correlation map.
    n_cols : int
        number of columns in the correlation map.

    Returns
    -------
    u : np.ndarray
        (n_rows, n_cols) array of x-direction velocimetry results in pixel displacements.
    v : np.ndarray
        (n_rows, n_cols) array of y-direction velocimetry results in pixel displacements.

    """
    u = np.zeros((n_rows, n_cols))
    v = np.zeros((n_rows, n_cols))

    # center point of the correlation map
    default_peak_position = np.floor(np.array(corr[0, :, :].shape) / 2)
    for k in nb.prange(n_rows):
        for m in nb.prange(n_cols):
            peak = (
                peak_position(
                    corr[k * n_cols + m],
                )
                - default_peak_position
            )
            u[k, m] = peak[1]
            v[k, m] = peak[0]
    return u, v


@nb.jit(cache=True, nopython=True)
def peak_position(corr):
    """Compute peak positions for correlations in each interrogation window using numba back-end."""
    eps = 1e-7
    idx = np.argmax(corr)
    peak1_i, peak1_j = idx // len(corr), idx % len(corr)
    # check if valid
    valid = peak1_i != 0 and peak1_i != corr.shape[-2] - 1 and peak1_j != 0 and peak1_j != corr.shape[-1] - 1
    if valid:
        corr = corr + eps  # prevents log(0) = nan if "gaussian" is used (notebook)
        c = corr[peak1_i, peak1_j] + eps
        cl = corr[peak1_i - 1, peak1_j] + eps
        cr = corr[peak1_i + 1, peak1_j] + eps
        cd = corr[peak1_i, peak1_j - 1] + eps
        cu = corr[peak1_i, peak1_j + 1] + eps

        # gaussian peak
        nom1 = np.log(cl) - np.log(cr)
        den1 = 2 * np.log(cl) - 4 * np.log(c) + 2 * np.log(cr) + eps
        nom2 = np.log(cd) - np.log(cu)
        den2 = 2 * np.log(cd) - 4 * np.log(c) + 2 * np.log(cu) + eps

        subp_peak_position = np.array([peak1_i + nom1 / den1, peak1_j + nom2 / den2])
    else:
        subp_peak_position = np.array([np.nan, np.nan])
    return subp_peak_position



"""tests for window manipulations."""

import numpy as np
import pytest

from ffpiv import window


def test_get_axis_shape(imgs):
    # get the last dimension (x-axis), assert if it fits
    dim_size = imgs.shape[-1]

    x_shape = window.get_axis_shape(
        dim_size=dim_size,
        window_size=64,
        overlap=32,
    )
    assert x_shape == 30


def test_get_array_shape(imgs):
    # get last two dimensions, assert numbers in returned dims
    dim_size = imgs.shape[-2:]
    xy_shape = window.get_array_shape(dim_size=dim_size, window_size=(64, 64), overlap=(32, 32))
    assert xy_shape == (36, 30)


@pytest.mark.parametrize(
    ("search_area_size", "test_coords", "len_coords"),
    [(64, [32.0, 64.0, 96.0, 128.0], 30), (128, [64.0, 96.0, 128.0, 160.0], 28)],
)
def test_get_axis_coords(imgs, search_area_size, test_coords, len_coords):
    dim_size = imgs.shape[-1]
    coords = window.get_axis_coords(dim_size, 64, 32, search_area_size=search_area_size)
    assert len(coords) == len_coords
    assert np.allclose(np.array(coords[0:4]), np.array(test_coords))


def test_get_rect_coordinates(imgs):
    x, y = window.get_rect_coordinates(
        dim_size=imgs.shape[-2:],
        window_size=(64, 64),
        overlap=(32, 32),
    )
    # test first block of coords
    assert len(y), len(x) == (11, 11)
    xi, yi = np.meshgrid(x, y)
    assert np.allclose(xi[0:2, 0:2], np.array([[32.0, 64], [32.0, 64.0]]))
    assert np.allclose(yi[0:2, 0:2], np.array([[32.0, 32.0], [64.0, 64.0]]))


@pytest.mark.parametrize("search_area_size", [(64, 64), (128, 128)])
def test_mask_search_area(search_area_size, window_size=(64, 64)):
    """Test creation of mask area."""
    mask = window.mask_search_area(window_size=window_size, search_area_size=search_area_size)
    assert mask.shape == search_area_size  # these must be the same
    assert mask.sum() == window_size[0] * window_size[1]  # amount of ones must be equal to the window_size surface
    start_idx_y = int((search_area_size[0] - window_size[0]) / 2)
    start_idx_x = int((search_area_size[1] - window_size[1]) / 2)
    assert np.allclose(mask[start_idx_y:-start_idx_y, start_idx_x:-start_idx_x], 1)


@pytest.mark.parametrize(("search_area_size", "test_win_size"), [((128, 128), 28 * 34), ((64, 64), 30 * 36)])
def test_sliding_window_array(imgs, search_area_size, test_win_size):
    win_x, win_y = window.sliding_window_idx(imgs[0], search_area_size=search_area_size)
    img_wins = window.sliding_window_array(imgs[0], win_x, win_y)
    assert img_wins.shape == (test_win_size, *search_area_size)


@pytest.mark.parametrize(
    ("swap_time_dim", "test_dims"),
    [(False, (4, 30 * 36, 64, 64)), (True, (30 * 36, 4, 64, 64))],
)
def test_multi_sliding_window_array(imgs, swap_time_dim, test_dims):
    # get the x and y coordinates per window
    win_x, win_y = window.sliding_window_idx(imgs[0])
    # apply the coordinates on all images
    window_stack = window.multi_sliding_window_array(imgs, win_x, win_y, swap_time_dim=swap_time_dim)
    assert window_stack.shape == test_dims


def test_normalize(imgs_win):
    img_norm = window.normalize(imgs_win, mode="xy")
    # check if shape remains the same
    assert imgs_win.shape == img_norm.shape
    img_norm = window.normalize(imgs_win, mode="time")
    assert imgs_win.shape == img_norm.shape



"""Numba just in time compiling cross-correlation related functions."""

import numba as nb
import numpy as np

from ffpiv.nb_utils import u_v_displacement


@nb.jit(cache=True, nopython=True)
def rfft2(x):
    """JIT variant of rfft2."""
    return np.fft.rfft2(x)


@nb.jit(cache=True, nopython=True)
def rfft2_x(x, fsize):
    """JIT variant of rfft2 with fsize parameter."""
    return np.fft.rfft2(x, fsize)


@nb.jit(cache=True, nopython=True)
def irfft2(x):
    """JIT variant of irfft2."""
    return np.fft.irfft2(x)


@nb.jit(cache=True, nopython=True)
def fftshift(x, axes):
    """JIT variant of fftshift."""
    return np.fft.fftshift(x, axes)


@nb.jit(cache=True, nopython=True)
def conj(x):
    """JIT variant of conj."""
    return np.conj(x)


@nb.jit(nb.float32[:, :](nb.float32[:, :]), cache=True, nopython=True)
def _normalize_intensity_window(img: np.ndarray) -> np.ndarray:
    """Normalize intensity of an image interrogation window using numba back-end.

    Parameters
    ----------
    img : np.ndarray (y, x)
        Image window

    Returns
    -------
    np.ndarray
        (y, x) array with normalized intensities of window

    """
    img_mean = np.float32(np.mean(img))
    img = img - img_mean
    img_std = np.float32(np.std(img))
    if img_std != 0:
        img = (img / img_std)
    else:
        img = np.zeros_like(img, dtype=np.float32)
    return img.astype(np.float32)

@nb.jit(nb.float32[:, :](nb.float32[:, :]), cache=True, nopython=True)
def _normalize_intensity_window_clip(img: np.ndarray) -> np.ndarray:
    """Normalize and clip (0, max value) intensity of an image interrogation window using numba back-end.

    Parameters
    ----------
    img : np.ndarray (y, x)
        Image window

    Returns
    -------
    np.ndarray
        (y, x) array with normalized intensities of window

    """
    img_mean = np.float32(np.mean(img))
    img = img - img_mean
    img_std = np.float32(np.std(img))
    if img_std != 0:
        img = img / img_std
    else:
        img = np.zeros_like(img, dtype=np.float32)
    return np.clip(img, 0, img.max())


@nb.jit(
    nb.float32[:, :, :](nb.float32[:, :, :], nb.float32[:, :, :], nb.boolean),
    parallel=True,
    nogil=True,
    cache=True,
    nopython=True
)
def _ncc(image_a, image_b, clip_norm):
    """Perform normalized cross correlation performed on a set of interrogation window pairs with numba back-end.

    Parameters
    ----------
    image_a : np.ndarray float32
        uint8 type array (w, y, x) containing a single image, sliced into interrogation windows (w)
    image_b : np.ndarray float32
        uint8 type array (w, y, x) containing the next image, sliced into interrogation windows (w)
    clip_norm: bool, optional
        If set to True, the normalized intensities are clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.

    Returns
    -------
    np.ndarray
        float64 (w, y, x) correlations of interrogation window pixels

    """
    res = np.empty_like(image_a, dtype=nb.float32)
    const = np.multiply(*image_a.shape[-2:])
    # below looks like duplication but is necessary for numba to compile to a fast running function
    if clip_norm:
        for n in nb.prange(image_a.shape[0]):
            ima = image_a[n]
            imb = image_b[n]
            ima = _normalize_intensity_window_clip(ima)
            imb = _normalize_intensity_window_clip(imb)
            f2a = conj(rfft2(ima))
            f2b = rfft2(imb)
            corr = fftshift(irfft2(f2a * f2b).real, axes=(-2, -1))
            res[n] = np.clip(corr / const, 0, 1)
    else:
        for n in nb.prange(image_a.shape[0]):
            ima = image_a[n]
            imb = image_b[n]
            ima = _normalize_intensity_window(ima)
            imb = _normalize_intensity_window(imb)
            f2a = conj(rfft2(ima))
            f2b = rfft2(imb)
            corr = fftshift(irfft2(f2a * f2b).real, axes=(-2, -1))
            res[n] = np.clip(corr / const, 0, 1)
    return res


@nb.jit(nogil=True, cache=True, nopython=True)
def _slice_a_b(imgs, n, mask, idx):
    """Extract one frame as source and the next as destination for image velocimetry.

    This function masks non-relevant areas in the source image and removes windows that are not relevant.

    Parameters
    ----------
    imgs : np.ndarray
        (i, w, y, x) set of images (i), subdivided into windows (w) for cross-correlation computation.
    n : int
        index of imgs to extract. n + 1 will be extracted as next frame
    mask : np.ndarray
        (y, x) array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.

    """
    return imgs[n, idx] * mask[idx], imgs[n + 1, idx]


@nb.jit(
    nb.float32[:, :, :, :](nb.float32[:, :, :, :], nb.float32[:, :, :], nb.boolean[:], nb.boolean),
    cache=True,
    parallel=True,
    nogil=True,
    nopython=True
)
def multi_img_ncc(imgs, mask, idx, clip_norm):
    """Compute correlation over all image pairs in `imgs` using numba back-end.

    Correlations are computed for each interrogation window (dim1) and each image pair (dim0)
    Because pair-wise correlation is performed the resulting dim0 size one stride smaller than the input imgs array.

    Parameters
    ----------
    imgs : np.ndarray float32
        (i, w, y, x) set of images (i), subdivided into windows (w) for cross-correlation computation.
    mask : np.ndarray float32
        (y, x) array containing ones in the area covered by a window, and zeros in the search area around the window.
    idx : np.ndarray, optional
        contains which windows (dimension w in imgs) should be cross correlated. If not provided, all windows are
        treated.
    clip_norm: bool, optional
        If set to True, the normalized intensities are clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.

    Returns
    -------
    np.ndarray
        float32 (i - 1, w, y, x) correlations of interrogation window pixels for each image pair spanning i.

    """
    corr = np.empty(
        shape=(len(imgs) - 1, imgs.shape[-3], imgs.shape[-2], imgs.shape[-1]),
        dtype=nb.float32,
    )
    corr.fill(np.nan)
    for n in nb.prange(len(imgs) - 1):
        img_a, img_b = _slice_a_b(imgs, n, mask, idx)
        corr[n, idx] = _ncc(img_a, img_b, clip_norm).astype(nb.float32)
    return corr


@nb.jit(parallel=True, cache=True, nopython=True)
def multi_u_v_displacement(
    corr,
    n_rows,
    n_cols,
):
    """Compute u and v displacement for multiple images at once.

    Parameters
    ----------
    corr : np.ndarray
        (i, w, y, x) correlations for each interrogation window (w) in each image pair (i).
    n_rows : int
        number of rows in end result
    n_cols : int
        number of columns in end result

    Returns
    -------
    u : np.ndarray
        Stack of x-direction velocimetry results (i, Y, X) in pixel displacements.
    v : np.ndarray
        Stack of y-direction velocimetry results (i, Y, X) in pixel displacements.

    """
    u = np.zeros((len(corr), n_rows, n_cols))
    v = np.zeros((len(corr), n_rows, n_cols))
    for i in nb.prange(len(corr)):
        _u, _v = u_v_displacement(corr[i], n_rows, n_cols)
        u[i] = _u
        v[i] = _v
    return u, v


@nb.jit(nb.float64[:](nb.float64[:, :, :]), parallel=True, cache=True, nogil=True, nopython=True)
def signal_to_noise(corr: np.ndarray):
    """Compute signal-to-noise ratio per interrogation window.

    This is computed by dividing the peak of the correlation field by the mean.

    Parameters
    ----------
    corr : np.ndarray
        (w, y, x) correlations for each interrogation window (w).

    Returns
    -------
    np.ndarray
        vector with signal to noise ratios for each window.

    """
    s2n = np.empty(
        len(corr),
        dtype=nb.float64,
    )
    for n in nb.prange(len(corr)):
        _c = corr[n]
        _c_max = _c.max()
        if _c_max < 1e-3:
            # no correlation
            s2n[n] = 0.0
        else:
            s2n[n] = _c_max / abs(_c.mean())
    return s2n


@nb.jit(nb.float64[:, :](nb.float64[:, :, :, :]), parallel=True, cache=True, nogil=True, nopython=True)
def multi_signal_to_noise(corr: np.ndarray):
    """Compute signal to noise for multiple images at once."""
    s2n = np.zeros(corr.shape[0:2])
    for i in nb.prange(len(corr)):
        _s2n = signal_to_noise(corr[i])
        s2n[i] = _s2n
    return s2n



import time

import numpy as np
import pytest

import ffpiv.nb_utils
from ffpiv import HAS_ROCKET_FFT

if HAS_ROCKET_FFT:
    import ffpiv.pnb as pnb

import ffpiv.pfftw as pfftw
import ffpiv.pnp as pnp
from ffpiv import window


@pytest.fixture()
def img_pair(imgs_win):
    # only return image 0 and 1
    img_pair = imgs_win[0:2]
    return img_pair


@pytest.fixture()
def dims(imgs):
    x, y = window.get_rect_coordinates(
        dim_size=imgs.shape[-2:],
        window_size=(64, 64),
        overlap=(32, 32),
    )
    nrows, ncols = len(y), len(x)
    return nrows, ncols


@pytest.fixture()
def correlations(img_pair):
    corrs = pfftw._ncc(*img_pair, clip_norm=False)
    return corrs * np.random.rand(*corrs.shape) * 0.005


@pytest.mark.parametrize("clip_norm", [False])
def test_ncc(img_pair, clip_norm):
    """Test correlation analysis on a pair of image windows."""
    image_a, image_b = img_pair

    img_1 = pfftw.normalize_intensity(image_a.astype(np.float32))
    img_2 = pnp.normalize_intensity(image_a.astype(np.float32))
    assert np.allclose(img_1, img_2)

    t1 = time.time()
    res_np = pnp.ncc(image_a, image_b, clip_norm)
    t2 = time.time()
    time_np = t2 - t1
    print(f"Numpy took {time_np} secs.")
    if HAS_ROCKET_FFT:
        t1 = time.time()
        res_nb = pnb._ncc(image_a, image_b, clip_norm)
        t2 = time.time()
        time_fftw = t2 - t1
        print(f"Numba took {time_fftw} secs.")
        assert np.allclose(res_nb, res_np, atol=1e-6, rtol=1e-5)
    t1 = time.time()
    res_fftw = pfftw._ncc(image_a, image_b, norm=True, clip_norm=False)
    t2 = time.time()
    time_nb = t2 - t1
    print(f"FFTW took {time_nb} secs.")
    assert np.allclose(res_fftw, res_np, atol=1e-6, rtol=1e-5)


def test_multi_img_ncc(imgs_win_stack, mask):
    """Test cross correlation with several hundreds of images."""
    t1 = time.time()
    res_np = pnp.multi_img_ncc(imgs_win_stack, mask, clip_norm=False)
    t2 = time.time()
    time_nb = t2 - t1
    print(f"Numpy took {time_nb} secs.")
    if HAS_ROCKET_FFT:
        t1 = time.time()
        idx = np.repeat(True, imgs_win_stack.shape[-3])
        res_nb = pnb.multi_img_ncc(imgs_win_stack, mask, idx, clip_norm=False)
        t2 = time.time()
        time_nb = t2 - t1
        print(f"Numba took {time_nb} secs.")
        assert np.allclose(res_nb, res_np, atol=1e-6, rtol=1e-5)
    t1 = time.time()
    res_fftw = pfftw.multi_img_ncc(imgs_win_stack, mask, clip_norm=False)
    t2 = time.time()
    time_fftw = t2 - t1
    print(f"FFTW took {time_fftw} secs.")
    assert np.allclose(res_fftw, res_np, atol=1e-6, rtol=1e-5)


def test_u_v_displacement(correlations, dims):
    """Test displacement functionalities."""
    if HAS_ROCKET_FFT:
        n_rows, n_cols = dims
        t1 = time.time()
        _ = ffpiv.nb_utils.u_v_displacement(correlations, n_rows, n_cols)
        t2 = time.time()
        print(f"Peak position search with numba took {t2 - t1} seconds")

    n_rows, n_cols = dims
    t1 = time.time()
    _ = pnp.u_v_displacement(correlations, n_rows, n_cols)
    t2 = time.time()
    print(f"Peak position search with numpy took {t2 - t1} seconds")

    # plt.quiver(u2, v2, color="r", alpha=0.5)
    # plt.quiver(u, v, color="b", alpha=0.5)
    # plt.show()


def test_peaks_numpy(correlations):
    peaks = pnp.peak_position(correlations)
    print(peaks)


@pytest.mark.skipif(not HAS_ROCKET_FFT, reason="Rocket-FFT not available, skipping signal to noise test.")
def test_signal_to_noise(correlations):
    # compile
    if HAS_ROCKET_FFT:
        _ = pnb.signal_to_noise(correlations)
        t1 = time.time()
        _ = pnb.signal_to_noise(correlations)
        t2 = time.time()
        print(f"Signal to noise calculation using Numba took {t2 - t1} seconds")



"""Interfacing wrapper functions to disclose functionalities."""

import gc
import time
import warnings
from typing import Literal, Optional, Tuple

import numpy as np

from ffpiv import window, HAS_ROCKET_FFT

if HAS_ROCKET_FFT:
    import ffpiv.pnb as pnb
import ffpiv.pfftw as pfftw
import ffpiv.pnp as pnp

def check_engine(engine: Literal["fftw", "numba", "numpy"] = "fftw"):
    """Check if the requested engine is available.

    If not, an alternative will be attempted to be selected and returned
    """
    if engine == "numba" and not HAS_ROCKET_FFT:
        warnings.warn('You selected "numba" but rocket_fft is not installed, switching to "fftw"', stacklevel=2)
    if engine == "numba" and HAS_ROCKET_FFT:
        return engine
    if engine == "fftw" or engine == "numpy":
        return engine
    else:
        return "fftw"


def subwindows(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    search_area_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
):
    """Subdivide image stack into windows with associated coordinates of center."""
    x, y = window.get_rect_coordinates(
        dim_size=imgs.shape[-2:], window_size=window_size, overlap=overlap, search_area_size=search_area_size
    )

    win_x, win_y = window.sliding_window_idx(
        imgs[0],
        window_size=window_size,
        search_area_size=search_area_size,
        overlap=overlap,
    )
    window_stack = window.multi_sliding_window_array(imgs, win_x, win_y)
    return x, y, window_stack


def coords(
    dim_size: Tuple[int, int],
    window_size: Tuple[int, int],
    overlap: Tuple[int, int],
    search_area_size: Optional[Tuple[int, int]] = None,
    center_on_field: bool = False,
):
    """Create coordinates (x, y) of velocimetry results.

    Overlap can be provided in case each interrogation window is to overlap with the neighbouring interrogation window.

    Parameters
    ----------
    dim_size : Tuple[int, int]
        size of the ingoing images (y, x)
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension.
    search_area_size : tuple[int, int], optional
        Search area window size in y (first) and x (second) dimension. If not provided, set to window_size
    center_on_field : bool, optional
        whether the center of interrogation window is returned (True) or (False) the bottom left (default=True)

    Returns
    -------
    x, y: np.ndarray (1D), np.ndarray (1D)
        x- and y-coordinates in axis form

    """
    return window.get_rect_coordinates(
        dim_size=dim_size,
        window_size=window_size,
        overlap=overlap,
        search_area_size=search_area_size,
        center_on_field=center_on_field,
    )


def piv(
    img_a: np.ndarray,
    img_b: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
    engine: Literal["numba", "numpy", "fftw"] = "fftw",
    clip_norm: bool = False,
):
    """Perform particle image velocimetry on a pair of images.

    Parameters
    ----------
    img_a : np.ndarray
        First image.
    img_b : np.ndarray
        Second image.
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension.
    engine : Literal["fftw", "numba", "numpy"], optional
        Compute correlations and displacements with "fftw" (default) or "numpy" or "numba"
        "numba" only works for python <= 3.12 as it depends on rocket_fft, which is unsupported on later versions.
    clip_norm : bool, optional
        If set to True, the normalized intensities is clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.


    Returns
    -------
    u : np.ndarray
        X-direction velocimetry results in pixel displacements.
    v : np.ndarray
        Y-direction velocimetry results in pixel displacements.

    """
    engine = check_engine(engine)
    # get subwindows
    imgs = np.stack((img_a, img_b), axis=0).astype(np.float64)
    # get correlations and row/column layout
    x, y, corr = cross_corr(imgs, window_size=window_size, overlap=overlap, engine=engine, clip_norm=clip_norm)
    # get displacements
    n_rows, n_cols = len(y), len(x)
    u, v = u_v_displacement(corr, n_rows, n_cols)
    return u[0], v[0]


def piv_stack(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (0, 0),
    engine: Literal["fftw", "numba", "numpy"] = "fftw",
    clip_norm: bool = False,
):
    """Perform particle image velocimetry over a stack of images.

    Parameters
    ----------
    imgs : np.ndarray
        Stack of images [i * y * x]
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x( second) dimension
    engine : Literal["fftw", "numba", "numpy"], optional
        Compute correlations and displacements with "fftw" (default) or "numpy" or "numba"
        "numba" only works for python <= 3.12 as it depends on rocket_fft, which is unsupported on later versions.
    clip_norm : bool, optional
        If set to True, the normalized intensities is clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.


    Returns
    -------
    u : np.ndarray
        Stack of x-direction velocimetry results (i -1, Y, X) in pixel displacements.
    v : np.ndarray
        Stack of y-direction velocimetry results (i -1, Y, X) in pixel displacements.

    """
    # get correlations and row/column layout
    t0 = time.time()
    x, y, corr = cross_corr(imgs, window_size=window_size, overlap=overlap, engine=engine, clip_norm=clip_norm)
    t1 = time.time()
    # get displacements
    n_rows, n_cols = len(y), len(x)
    print(f"UV displacements calculating after {t1 -t0} s.")
    if engine == "numpy":
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    else:
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    return u, v


def cross_corr(
    imgs: np.ndarray,
    window_size: Tuple[int, int] = (64, 64),
    overlap: Tuple[int, int] = (32, 32),
    search_area_size: Optional[Tuple[int, int]] = None,
    engine: Literal["fftw", "numba", "numpy"] = "fftw",
    normalize: bool = False,
    clip_norm: bool = False,
    verbose: bool = True,
):
    """Compute correlations over a stack of images using interrogation windows.

    Parameters
    ----------
    imgs : np.ndarray
        Stack of images [i * y * x]
    window_size : tuple[int, int], optional
        Interrogation window size in y (first) and x (second) dimension.
    overlap : tuple[int, int], optional
        Overlap on window sizes in y (first) and x (second) dimension.
    search_area_size : tuple[int, int], optional
        size of the search area. This is used in the second frame window set, to explore areas larger than
        `window_size`.
    engine : Literal["fftw", "numba", "numpy"], optional
        The engine to use for calculation, by default "fftw".
        "numba" only works for python <= 3.12 as it depends on rocket_fft, which is unsupported on later versions.
    normalize : bool, optional
        if set, each window will be normalized with spatial mean and standard deviation, and numbers capped to 0.
    clip_norm : bool, optional
        If set to True, the normalized intensities is clipped to the range [0, max] where max is the maximum of the
        window, before FFT is performed.
    verbose : bool, optional
        if set (default), warnings will be displayed if the amount of available memory is low.

    Returns
    -------
    n_rows : int
        The number of rows of windows.
    n_cols : int
        The number of columns of windows.
    corr : np.ndarray
        A 4D array containing per image and per interrogation window the correlation results.

    """
    engine = check_engine(engine)
    # check if masking is needed
    window_size = window.round_to_even(window_size)
    if search_area_size is None:
        search_area_size = window_size
    # check if search_area_size contains uneven numbers
    search_area_size = window.round_to_even(search_area_size)
    # search_area_size must be at least equal to the window size
    search_area_size = max(search_area_size, window_size)

    # estimate amount of memory required for intermediate in-memory storage
    dim_size = (imgs.shape[-2], imgs.shape[-1])
    # win_shape = window.get_array_shape(
    #     dim_size=(imgs.shape[-2], imgs.shape[-1]),
    #     window_size=window_size,
    #     overlap=overlap,
    #     search_area_size=search_area_size
    # )
    # var_size1 = len(imgs) * win_shape[0] * win_shape[1] * search_area_size[0] * search_area_size[1] * 4 / 1e9
    # Prepare subwindows
    imgs = np.float32(imgs)
    req_mem = window.required_memory(
        len(imgs), dim_size=dim_size, window_size=window_size, overlap=overlap, search_area_size=search_area_size
    )
    avail_mem = window.available_memory()
    if verbose:
        if avail_mem - req_mem < 0:
            warnings.warn(
                f"You may have too little physical memory ({avail_mem / 1e9} GB) available for this problem. "
                f"You may need {req_mem / 1e9} GB. ffpiv may slow down or crash! Reduce the amount of frames "
                f" interpreted in one go.",
                stacklevel=2,
            )
            # wait for a while so that user can read the message
            time.sleep(1)
    x, y, window_stack = subwindows(
        imgs,
        window_size=window_size,
        search_area_size=search_area_size,
        overlap=overlap,
    )
    # normalization
    if normalize:
        window_stack = window.normalize(window_stack, mode="xy")
    # prepare a mask for the first frame of analysis
    mask = window.mask_search_area(window_size=window_size, search_area_size=search_area_size)
    # expand mask over total amount of sub windows
    mask = np.repeat(np.expand_dims(mask, 0), window_stack.shape[1], axis=0).astype(np.float32)

    # fully missing should be ignored
    idx = np.any(window_stack[0] != 0, axis=(-1, -2))
    if engine == "numpy":
        corr = pnp.multi_img_ncc(window_stack, mask=mask, idx=idx, clip_norm=clip_norm)
    elif engine == "fftw":
        corr = pfftw.multi_img_ncc(window_stack, mask=mask, idx=idx, clip_norm=clip_norm)
    else:
        corr = pnb.multi_img_ncc(window_stack, mask=mask, idx=idx, clip_norm=clip_norm)
    # memory cleanup
    del idx, mask, window_stack
    gc.collect()
    return x, y, corr


def u_v_displacement(corr: np.array, n_rows: int, n_cols: int, engine: Literal["fftw", "numba", "numpy"] = "fftw"):
    """Compute x-direction and y-directional displacements.

    Parameters
    ----------
    corr : np.array
        4D array [i, w, x, y] cross correlations computed per image and interrogation window pixel.
    n_rows : int
        The number of rows in the output displacement arrays.
    n_cols : int
        The number of columns in the output displacement arrays.
    engine : Literal["fftw", "numba", "numpy"], optional
        The computational engine to use for calculating displacements, "fftw", "numba" or "numpy". Default is "fftw".
        "numba" only works for python <= 3.12 as it depends on rocket_fft, which is unsupported on later versions.

    Returns
    -------
    u : np.array
        An array containing the u-component of the displacements.
    v : np.array
        An array containing the v-component of the displacements.

    """
    engine = check_engine(engine)
    if engine == "numpy":
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    else:
        # from ffpiv.pnb import multi_u_v_displacement
        u, v = pnp.multi_u_v_displacement(corr, n_rows, n_cols)
    return u, v



"""Retrieval of sample dataset."""

import glob
import os.path
import zipfile


def get_hommerich_dataset():
    """Retrieve and cache sample dataset of Sheaf river."""
    try:
        import pooch
    except ImportError:
        raise ImportError("This function needs pooch. Install ffpiv with pip install ffpiv[extra]")

    # Define the DOI link
    filename = "hommerich_frames_20241010_081717.zip"
    base_url = "https://zenodo.org/records/15002591/files"
    url = base_url + "/" + filename
    print(f"Retrieving or providing cached version of dataset from {url}")
    # Create a Pooch registry to manage downloads
    registry = pooch.create(
        # Define the cache directory
        path=pooch.os_cache("ffpiv"),
        # Define the base URL for fetching data
        base_url=base_url,
        # Define the registry with the file we're expecting to download
        registry={filename: None},
    )
    # Fetch the dataset
    file_path = registry.fetch(filename, progressbar=True)
    print(f"Hommerich dataset is available in {file_path}")
    return file_path


def get_hommerich_files():
    """Unzip hommerich dataset and return file list."""
    zip_file = get_hommerich_dataset()
    trg_dir = os.path.split(zip_file)[0]
    trg_subdir = os.path.join(trg_dir, "hommerich_frames_20241010_081717")
    jpg_template = os.path.join(trg_subdir, "frame*.jpg")
    if not (len(glob.glob(jpg_template)) >= 122):
        # apparently the jpg files are incomplete or not present. Unzip the record.
        print("Unzipping sample data...")
        with zipfile.ZipFile(zip_file, "r") as f:
            f.extractall(trg_dir)
    # make a neat sorted list of jpg files
    jpg_files = glob.glob(os.path.join(trg_dir, "hommerich_frames_20241010_081717", "frame*.jpg"))
    jpg_files.sort()
    return jpg_files
